#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
This module concerns the Fetcher facilities. There are expected to be about
three of these:

 - NgasDirectCopyFetcher, which uses the NGAS direct copy plugin,
 - NgasStreamingFileFetcher, which uses the NGAS streaming mode, and
 - OracleXmlFetcher, which queries the ALMA database to fetch something

Each of these is fairly stupid and just tries to fetch its one file, one time.
Retrying is not handled here, the RetryableFileFetcher is used for that.
"""
from pathlib import Path

import cx_Oracle
import sqlalchemy
from pycapo import CapoConfig

from .exceptions import RetryableFetchError, UnsatisfiableConstraint
from .interfaces import FetchContext, FileFetcher, LocatedFile
from .locations import NgasFile, OracleXml
from .retry import retry

# pylint: disable=E0401, E0402, W0221


class NgasDirectCopyFileFetcher(FileFetcher):
    """
    Fetch a file using the NGAS direct copy mechanism.
    """

    def __init__(self, file: NgasFile):
        super().__init__()
        self._file = file

    @property
    def file(self) -> NgasFile:
        return self._file

    def shares_resources_with(self, other: FileFetcher) -> bool:
        return self.file.shares_resources_with(other.file)

    def do_fetch(self, context: FetchContext) -> Path:
        """
        Fetch a file from NGAS via direct copy, or die trying.

        :return: the fetched file
        """
        # ensure that we actually can fetch
        if not self.file.can_direct_copy(context):
            raise UnsatisfiableConstraint(f"Direct copy to {self.destination(context)} is not possible")

        # check the destination
        destination = context.check_and_prepare_destination(self.file)

        # direct copy the file
        self.file.direct_copy(context)

        return destination

    def __str__(self):
        return f"NGAS Direct Copy of {self.file}"


class NgasStreamingFileFetcher(FileFetcher):
    """
    Fetch a file using NGAS streaming.
    """

    def __init__(self, file: NgasFile):
        super().__init__()
        self._file = file

    @property
    def file(self) -> NgasFile:
        return self._file

    def shares_resources_with(self, other: FileFetcher) -> bool:
        return self.file.shares_resources_with(other.file)

    def do_fetch(self, context: FetchContext) -> Path:
        """
        Fetch a file from NGAS via streaming, or die trying.

        :return: the fetched file
        """
        # check the destination
        destination = context.check_and_prepare_destination(self.file)

        # run the streaming fetch
        self.file.stream(context)

        return destination

    def __str__(self):
        return f"NGAS streaming copy of {self.file}"


class NgasModeDetectingFetcher(FileFetcher):
    """
    This fetcher is used when the user has no preference between streaming and direct-copy.
    The fetcher itself looks at the file and the destination and makes a determination about
    which one to use.
    """

    @property
    def file(self) -> NgasFile:
        return self._file

    def shares_resources_with(self, other: FileFetcher) -> bool:
        return self.file.shares_resources_with(other.file)

    def __init__(self, file: NgasFile):
        super().__init__()
        self._file = file

    def do_fetch(self, context: FetchContext) -> Path:
        # check the destination
        destination = context.check_and_prepare_destination(self.file)

        # either copy or stream
        if self.file.can_direct_copy(context):
            self.file.direct_copy(context)
        else:
            self.file.stream(context)

        # return the destination
        return destination


class OracleXmlFetcher(FileFetcher):
    """
    Fetch a file by retrieving it from the ALMA database.
    """

    ALMA_CAPO_PREFIX = "almaMetadataDatabase.jdbc"

    engine = None

    def __init__(self, file: OracleXml):
        super().__init__()
        self._file = file

    @property
    def file(self) -> OracleXml:
        return self._file

    def shares_resources_with(self, other: FileFetcher) -> bool:
        return self.file.shares_resources_with(other.file)

    def do_fetch(self, context: FetchContext) -> Path:
        """
        Retrieve an ALMA SDM table as XML from the ALMA DB.

        :return: the file we got
        """
        # check the destination
        destination = context.check_and_prepare_destination(self.file)

        engine = self.get_engine()
        try:
            with engine.connect() as conn:
                result = list(
                    conn.execute(
                        f"SELECT (xml).getClobVal() FROM ALMA.{self.file.table} WHERE " f"archive_uid = :archive_uid",
                        dict(archive_uid=self.file.archive_uid),
                    )
                )[0][0]

                destination.write_text(result)
                return destination
        except Exception as ex:
            raise RetryableFetchError(f"Oracle XML retrieval failed for {self.file}") from ex
        finally:
            engine.dispose()

    @classmethod
    def get_engine(cls):
        """
        Get an engine against which queries can be run.

        :return:
        """
        if cls.engine is None:
            capo_config = CapoConfig()
            user = capo_config[cls.ALMA_CAPO_PREFIX + "Username"]
            passwd = capo_config[cls.ALMA_CAPO_PREFIX + "Password"]
            url = capo_config[cls.ALMA_CAPO_PREFIX + "Url"]

            hostport, service_name = url.split("/")[-2:]
            host, port = hostport.split(":")
            dsn = cx_Oracle.makedsn(host, port, service_name=service_name)
            archiveurl = f"oracle://{user}:{passwd}@{dsn}"

            cls.engine = sqlalchemy.create_engine(archiveurl)
        return cls.engine

    def __str__(self):
        return f"Oracle XML fetch of {self.file}"


class FileFetcherDecorator(FileFetcher):
    """
    Decorator to allow us to interpose other behavior into a FileFetcher.
    """

    def __init__(self, underlying: FileFetcher):
        super().__init__()
        self.underlying = underlying

    @property
    def file(self) -> LocatedFile:
        return self.underlying.file

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        return self.underlying.shares_resources_with(other)

    def do_fetch(self, context: FetchContext) -> Path:
        return self.underlying.do_fetch(context)

    def __str__(self):
        return str(self.underlying)


class RetryableFileFetcher(FileFetcherDecorator):
    """A FileFetcher that retries N times in the event of failure, waiting
    longer each time"""

    def __init__(self, underlying: FileFetcher, retries=3):
        super().__init__(underlying)
        self.retries = retries

    def do_fetch(self, context: FetchContext) -> Path:
        return retry(lambda: self.underlying.do_fetch(context), [2**attempt for attempt in range(1, self.retries)])

    def __str__(self):
        return f"{self.underlying} (with up to {self.retries} retries)"


class DryRunFakeFileFetcher(FileFetcher):
    """
    A fake fetcher we can use to implement dry runs and other tests
    """

    def __init__(self, file: LocatedFile):
        super().__init__()
        self._file = file

    @property
    def file(self) -> LocatedFile:
        return self._file

    def do_fetch(self, context: FetchContext) -> Path:
        """
        In a dry run, we don't actually do any fetching.
        """
        return context.path_to(self.file)

    def shares_resources_with(self, other: FileFetcher) -> bool:
        return self.file.shares_resources_with(other.file)

    def __str__(self):
        return f"Fake fetch of {self.file} to {self.destination}"
